import os
import shutil
import pytest

from munch import Munch
from mock import patch, Mock, call
from tempfile import mkdtemp

from . import base

from rpkglib import GitClient
from rpkglib.utils import run
from rpkglib.errors import RpkgError

def silent_run(cmd, **kwargs):
    return run(cmd, **kwargs, capture_stderr=True)

class TestGitClient(base.TestCase):

    def setUp(self):
        super(TestGitClient, self).setUp()

    def tearDown(self):
        super(TestGitClient, self).tearDown()

    def test_git_props(self):
        self.init_repo()
        result = GitClient.git_props()
        head = self.repo_head()
        head_short = self.repo_head_short()
        self.assertEqual(result, Munch({
            'root': self.cwd,
            'branch': 'master',
            'remote': 'origin',
            'remote_url': 'https://pagure.io/rpkg-util',
            'remote_netloc': 'pagure.io',
            'head': head,
            'head_short': head_short,
            'status': '',
            'merged_tag_refs': '',
            'submodule_refs': '{} {}'.format(head, self.cwd),
            'dirty': '',
        }))

    @patch('rpkglib.GitClient.git_props', lambda: True)
    @patch('rpkglib.git_client.run')
    def test_write_ignores(self, run):
        git_client = GitClient(self.cwd)
        self.assertNotIn('.gitignore', os.listdir('.'))

        git_client.write_ignores(['filename1', 'filename2'])
        self.assertIn('.gitignore', os.listdir('.'))
        with open('.gitignore', 'r') as gitignore_file:
            self.assertEqual(
                [l.strip() for l in gitignore_file.readlines()],
                ['/filename1', '/filename2'])

        git_client.write_ignores(['filename1'])
        with open('.gitignore', 'r') as gitignore_file:
            self.assertEqual(
                [l.strip() for l in gitignore_file.readlines()],
                ['/filename1', '/filename2'])

        git_client.write_ignores(['filename1'], append=False)
        with open('.gitignore', 'r') as gitignore_file:
            self.assertEqual(
                [l.strip() for l in gitignore_file.readlines()],
                ['/filename1'])

    @patch('rpkglib.GitClient.git_props', lambda: True)
    @patch('rpkglib.git_client.run')
    def test_add_to_index(self, run):
        git_client = GitClient(self.cwd)
        git_client.add_to_index('file1', 'file2')
        run.assert_called_with(['git', 'add', 'file1', 'file2'])

    @patch('rpkglib.GitClient.git_props', lambda: True)
    @patch('rpkglib.git_client.run')
    def test_show_diff_index(self, run):
        git_client = GitClient(self.cwd)
        git_client.show_diff_index()
        run.assert_called_with(['git', 'diff', '--cached'], capture=False)

    @patch('rpkglib.GitClient.git_props', lambda: True)
    @patch('rpkglib.git_client.run')
    def test_ls_tree(self, run):
        git_client = GitClient(self.cwd)
        git_client.ls_tree('--name-only', 'HEAD')
        run.assert_called_with(['git', 'ls-tree', '--name-only', 'HEAD'])

    @patch('rpkglib.GitClient.git_props', lambda: True)
    @patch('rpkglib.git_client.run')
    def test_commit_id_from_tag_id(self, run):
        git_client = GitClient(self.cwd)
        git_client.commit_id_from_tag_id('tag_id')
        run.assert_called_with(['git', 'rev-parse', 'tag_id^{commit}'])

    def test_read_file(self):
        self.init_repo()
        run(['echo 1 > file'], shell=True)
        run(['git add file'], shell=True)
        run(['git commit file -m "new commit"'], shell=True)
        git_client = GitClient(self.cwd)
        result = git_client.read_file('HEAD', os.path.join(self.cwd, 'file'))
        self.assertEqual('1', result)

        result = git_client.read_file('HEAD', 'file')
        self.assertEqual('1', result)

        result = git_client.read_file('HEAD', '../'+os.path.basename(self.cwd)+'/file')
        self.assertEqual('1', result)

        with patch('rpkglib.git_client.run', silent_run):
            with pytest.raises(RpkgError):
                git_client.read_file('HEAD', os.path.join(self.cwd, 'filex'))

        with patch('rpkglib.git_client.run', silent_run):
            with pytest.raises(RpkgError):
                git_client.read_file('HEAD', os.path.join(self.cwd, '../file'))

    @patch('rpkglib.GitClient.git_props', lambda: True)
    @patch('rpkglib.git_client.run')
    def test_clone_url(self, run):
        git_client = GitClient(self.cwd,
                               clone_url='ssh://git@pagure.io/%(repo_path)s',
                               anon_clone_url='https://pagure.io/%(repo_path)s')

        result = git_client.clone_url(anonymous=False, repo_path='foo/bar')
        self.assertEqual(result, 'ssh://git@pagure.io/foo/bar')

        result = git_client.clone_url(anonymous=True, repo_path='foo/bar')
        self.assertEqual(result, 'https://pagure.io/foo/bar')

        result = git_client.clone_url(anonymous=True, repo_path='bar')
        self.assertEqual(result, 'https://pagure.io/bar')

        git_client = GitClient(self.cwd,
                               clone_url='ssh://git@pagure.io/%(repo_path)s',
                               anon_clone_url='https://pagure.io/%(repo_path)s',
                               default_namespace='rpms')

        result = git_client.clone_url(anonymous=True, repo_path='bar/')
        self.assertEqual(result, 'https://pagure.io/rpms/bar')

    def test_latest_merged_tag_id(self):
        self.init_repo()
        cmds = [
            ['git', 'tag', '-a', '-m', 'empty', 'x-1-1'],
            ['git', 'branch', 'alt'],
            ['touch', 'file'],
            ['git', 'add', 'file'],
            ['git', 'commit', '-m', 'new file'],
            ['git', 'tag', '-a', '-m', 'empty', 'x-2-1'],
            ['git', 'checkout', 'alt'],
            ['touch', 'altfile'],
            ['git', 'add', 'altfile'],
            ['git', 'commit', '-m', 'new altfile'],
            ['git', 'tag', '-a', '-m', 'empty', 'x-3-1'],
            ['git', 'tag', '-a', '-m', 'empty', 'y-1-1'],
            ['git', 'checkout', 'master'],
        ] # TODO: might require sleep between tags

        for cmd in cmds:
            silent_run(cmd)

        git_client = GitClient(self.cwd)
        result = git_client.latest_merged_tag_id('y')
        self.assertEqual(result, None)

        result = git_client.latest_merged_tag_id('x')
        expected_tag_id = self.tag_id('x-2-1')
        self.assertEqual(result, expected_tag_id)

        silent_run(['git', 'checkout', 'alt'])
        git_client_alt = GitClient(self.cwd)

        result = git_client_alt.latest_merged_tag_id('x')
        expected_tag_id = self.tag_id('x-3-1')
        self.assertEqual(result, expected_tag_id)

        result = git_client_alt.latest_merged_tag_id('y')
        expected_tag_id = self.tag_id('y-1-1')
        self.assertEqual(result, expected_tag_id)

    @patch('rpkglib.GitClient.git_props', lambda: Munch({'head': 'head_id', 'branch': 'branch'}))
    @patch('rpkglib.git_client.run')
    @patch('rpkglib.git_client.utils.edit')
    @patch('rpkglib.git_client.utils.changelog_entry')
    @patch('rpkglib.GitClient.latest_merged_tag_id')
    @patch('rpkglib.GitClient.commit_id_from_tag_id')
    def test_tag_message(self, commit_id_from_tag_id, latest_merged_tag_id,
                         changelog_entry, edit, run):
        commit_id_from_tag_id.return_value = 'commit_id'
        changelog_entry.return_value = 'Hello world!'
        edit.return_value = 'Hello world again!'
        run.return_value = Mock(out='')
        git_client = GitClient(self.cwd)
        result = git_client.tag_message('x-2-1')
        self.assertEqual(result, 'Commits: commit_id..head_id\n\nHello world again!')
        run.assert_has_calls([
            call(['git', 'log', '--format=%s', 'commit_id..head_id', self.cwd]),
            call(['git', 'var', 'GIT_EDITOR']),
        ], any_order=True)

        edit('', "Hello world!\n#\n# Write a changelog for tag:\n#   "
                 "x-2-1\n# Lines starting with '#' will be ignored.\n")

        run.reset_mock()
        edit.reset_mock()
        latest_merged_tag_id.return_value = None
        result = git_client.tag_message('x-2-1', skip_edit=True)
        self.assertEqual(result, "Commits: head_id\n\nHello world!\n#\n# Write a "
                                  "changelog for tag:\n#   x-2-1\n# Lines starting "
                                  "with '#' will be ignored.\n")
        run.assert_has_calls([
            call(['git', 'log', '--format=%s', 'head_id', self.cwd]),
            call(['git', 'var', 'GIT_EDITOR']),
        ], any_order=True)
        edit.assert_not_called()

    @patch('rpkglib.GitClient.git_props')
    def test_repo_path(self, git_props):
        git_props.return_value = Munch(remote_url='https://pagure.io/rpkg-util')
        git_client = GitClient(self.cwd,
                               clone_url='ssh://git@pagure.io/%(repo_path)s',
                               anon_clone_url='https://pagure.io/%(repo_path)s')

        result = git_client.repo_path()
        self.assertEqual(result, 'rpkg-util')

        git_props.return_value = Munch(remote_url='ssh://git@pagure.io/rpkg-util.git')
        git_client = GitClient(self.cwd,
                               clone_url='ssh://git@pagure.io/%(repo_path)s',
                               anon_clone_url='https://pagure.io/%(repo_path)s')
        result = git_client.repo_path()
        self.assertEqual(result, 'rpkg-util')

        git_props.return_value = Munch(remote_url='ssh://git@pagurx.io/foo/bar.git')
        git_client = GitClient(self.cwd,
                               clone_url='ssh://git@pagure.io/%(repo_path)s',
                               anon_clone_url='https://pagure.io/%(repo_path)s')
        result = git_client.repo_path()
        self.assertEqual(result, 'foo/bar')

        git_props.return_value = Munch(remote_url='ssh://git@pagurx.io/foo/bar.git')
        git_client = GitClient(self.cwd,
                               clone_url='ssh://git@pagurx.io/foo/%(repo_path)s',
                               anon_clone_url='https://pagure.io/%(repo_path)s')
        result = git_client.repo_path()
        self.assertEqual(result, 'bar')

    @patch('rpkglib.GitClient.git_props', lambda: True)
    @patch('rpkglib.git_client.run')
    def test_add(self, run):
        git_client = GitClient(self.cwd)

        git_client.add(False, ['path1', 'path2'])
        run.assert_called_with(['git', 'add', '--', 'path1', 'path2'],
                               capture=False)

        git_client.add(True, ['path1', 'path2'])
        run.assert_called_with(['git', 'add', '--all', '--', 'path1', 'path2'],
                               capture=False)

    @patch('rpkglib.GitClient.git_props', lambda: True)
    @patch('rpkglib.git_client.run')
    def test_checkout(self, run):
        git_client = GitClient(self.cwd)
        git_client.checkout('target', False)
        run.assert_called_with(['git', 'checkout', 'target'],
                               capture=False)

        git_client.checkout('target', True)
        run.assert_called_with(['git', 'checkout', '-b', 'target'],
                               capture=False)

    @patch('rpkglib.GitClient.git_props', lambda: True)
    @patch('rpkglib.git_client.run')
    def test_clean(self, run):
        git_client = GitClient(self.cwd)

        git_client.clean(dry_run=False, x=False)
        run.assert_called_with(['git', 'clean', '-f',
                                '-d'], capture=False)

        git_client.clean(dry_run=True, x=True)
        run.assert_called_with(['git', 'clean', '-n', '-x',
                                '-f', '-d'], capture=False)

        git_client = GitClient(self.cwd, clean_force=False,
                               clean_dirs=False)
        git_client.clean()
        run.assert_called_with(['git', 'clean'],
                               capture=False)

    @patch('rpkglib.GitClient.git_props', lambda: True)
    @patch('rpkglib.GitClient.clone_url')
    @patch('rpkglib.git_client.os.makedirs')
    @patch('rpkglib.git_client.run')
    def test_clone(self, run, makedirs, clone_url):
        git_client = GitClient(self.cwd)
        clone_url.return_value = 'https://pagure.io/rpkg-util'

        git_client.clone('rpkg-util', 'thisdir', branch=None)
        run.assert_called_with(['git', 'clone', 'https://pagure.io/rpkg-util',
                                'thisdir'], capture=False)

        git_client.clone('rpkg-util', 'thisdir', branch='foo')
        run.assert_called_with(['git', 'clone', 'https://pagure.io/rpkg-util',
                                'thisdir', '--branch', 'foo'], capture=False)

    @patch('rpkglib.GitClient.git_props', lambda: True)
    @patch('rpkglib.GitClient.clone_url')
    @patch('rpkglib.git_client.run', silent_run)
    @patch('rpkglib.git_client.os.makedirs')
    def test_clone_with_dirs(self, makedirs, clone_url):
        tmpdir = mkdtemp()
        self.init_bare_multibranch_repo()
        git_client = GitClient(self.cwd)
        clone_url.return_value = self.cwd

        git_client.clone_with_dirs('', tmpdir)

        ls_result = run(['ls'], cwd=tmpdir).out.split()
        self.assertEqual(ls_result, ['alt', 'master'])

        master_remote_result = run(
            'git config remote.$(git config branch.master.remote).url',
            shell=True, cwd=os.path.join(tmpdir, 'master')).out
        self.assertEqual(master_remote_result, self.cwd)

        alt_remote_result = run(
            'git config remote.$(git config branch.alt.remote).url',
            shell=True, cwd=os.path.join(tmpdir, 'alt')).out
        self.assertEqual(alt_remote_result, self.cwd)

        shutil.rmtree(tmpdir)

    @patch('rpkglib.GitClient.git_props', lambda: True)
    @patch('rpkglib.git_client.run')
    def test_commit(self, run):
        git_client = GitClient(self.cwd)
        git_client.commit()
        run.assert_called_with(['git', 'commit'], capture=False)

        git_client.commit(msg_lines=['a', 'b'])
        run.assert_called_with(['git', 'commit', '-m', 'a', '-m', 'b'],
                               capture=False)

        git_client.commit(all=True, amend=True, msg_lines=['a', 'b'])
        run.assert_called_with(['git', 'commit', '-a' , '--amend', '-m', 'a', '-m', 'b'],
                               capture=False)

        git_client.commit(msg_lines=[], msg_file='msgfile',
                          sign_off=True, files=['somefile'])
        run.assert_called_with(['git', 'commit', '-F', 'msgfile', '-s',
                                '--', 'somefile'], capture=False)

        git_client = GitClient(self.cwd, gpg_keyid='abc')
        git_client.commit()
        run.assert_called_with(['git', 'commit', '-Sabc'], capture=False)

    @patch('rpkglib.GitClient.git_props', lambda: True)
    @patch('rpkglib.git_client.run')
    def test_diff(self, run):
        git_client = GitClient(self.cwd)
        git_client.diff()
        run.assert_called_with(['git', 'diff'], capture=False)

        git_client.diff(True, ['somefile1', 'somefile2'])
        run.assert_called_with(['git', 'diff', '--cached',
                                '--', 'somefile1', 'somefile2'], capture=False)

    @patch('rpkglib.GitClient.git_props', lambda: True)
    @patch('rpkglib.git_client.run')
    def test_fetch(self, run):
        git_client = GitClient(self.cwd)
        git_client.fetch()
        run.assert_called_with(['git', 'fetch'],
                               capture=False)

    @patch('rpkglib.GitClient.git_props', lambda: True)
    @patch('rpkglib.git_client.run')
    def test_list_branches(self, run):
        git_client = GitClient(self.cwd)
        git_client.list_branches()
        run.assert_called_with(['git', 'branch', '-a'],
                               capture=False)

    @patch('rpkglib.GitClient.git_props', lambda: True)
    @patch('rpkglib.git_client.run')
    def test_log(self, run):
        git_client = GitClient(self.cwd)

        git_client.log()
        run.assert_called_with(['git', 'log', '--graph', '--all'], capture=False)

        git_client.log(show_signature=True, positionals=['HEAD', '.', '../'])
        run.assert_called_with(['git', 'log', '--graph', '--all', '--show-signature',
                                'HEAD', '.', '../'], capture=False)

        git_client.log(show_signature=True, positionals=['HEAD', '--', '.'])
        run.assert_called_with(['git', 'log', '--graph', '--all', '--show-signature',
                                'HEAD', '--', '.'], capture=False)

        git_client = GitClient(self.cwd, log_graph=False, log_all=False)

        git_client.log()
        run.assert_called_with(['git', 'log'], capture=False)

        git_client.log(graph=True, all=True)
        run.assert_called_with(['git', 'log', '--graph', '--all'], capture=False)

    @patch('rpkglib.GitClient.git_props', lambda: True)
    @patch('rpkglib.git_client.run')
    def test_merge_continue(self, run):
        git_client = GitClient(self.cwd)
        git_client.merge_continue()
        run.assert_called_with(['git', 'merge', '--continue'], capture=False)

    @patch('rpkglib.GitClient.git_props', lambda: True)
    @patch('rpkglib.git_client.run')
    def test_merge_abort(self, run):
        git_client = GitClient(self.cwd)
        git_client.merge_abort()
        run.assert_called_with(['git', 'merge', '--abort'], capture=False)

    @patch('rpkglib.GitClient.git_props', lambda: True)
    @patch('rpkglib.git_client.run')
    def test_merge(self, run):
        git_client = GitClient(self.cwd)
        git_client.merge()
        run.assert_called_with(['git', 'merge'], capture=False)

        run.reset_mock()
        git_client = GitClient(self.cwd, gpg_keyid='abc')
        git_client.merge(no_commit=True, squash=True,
                         message="msg", msg_file="msg_file",
                         sign_off=True, commits=["commit1", "commit2"])
        run.assert_called_with(['git', 'merge', '--no-commit', '--squash',
                                '-m', 'msg', '-F', 'msg_file', '--signoff',
                                '-Sabc', '--', 'commit1', 'commit2'], capture=False)

    @patch('rpkglib.GitClient.git_props', lambda: True)
    @patch('rpkglib.GitClient.latest_merged_tag_id')
    @patch('rpkglib.git_client.run')
    def test_new(self, run, latest_merged_tag_id):
        latest_merged_tag_id.return_value = 'someid'
        git_client = GitClient(self.cwd)
        git_client.new('x')
        run.assert_called_with(['git', 'diff', 'someid'], capture=False)


    @patch('rpkglib.GitClient.git_props', lambda: True)
    @patch('rpkglib.git_client.run')
    def test_pull(self, run):
        git_client = GitClient(self.cwd)
        git_client.pull()
        run.assert_called_with(['git', 'pull'], capture=False)

        git_client.pull(True)
        run.assert_called_with(['git', 'pull', '--rebase'],
                               capture=False)

        git_client.pull(False, True)
        run.assert_called_with(['git', 'pull', '--no-rebase'],
                               capture=False)

    @patch('rpkglib.GitClient.git_props', lambda: True)
    @patch('rpkglib.git_client.run')
    def test_push(self, run):
        git_client = GitClient(self.cwd)
        git_client.push()
        run.assert_called_with(['git', 'push', '--follow-tags'],
                               capture=False)

        git_client.push(True)
        run.assert_called_with(['git', 'push', '-f', '--follow-tags'],
                               capture=False)

        git_client = GitClient(self.cwd, push_follow_tags=False)
        git_client.push()
        run.assert_called_with(['git', 'push'],
                               capture=False)

    @patch('rpkglib.GitClient.git_props', lambda: True)
    @patch('rpkglib.git_client.run')
    def test_reset(self, run):
        git_client = GitClient(self.cwd)
        git_client.reset()
        run.assert_called_with(['git', 'reset'],
                               capture=False)

        git_client.reset(1, 'path')
        run.assert_called_with(['git', 'reset', '--mixed', 'path'],
                               capture=False)

        git_client.reset(2, 'path')
        run.assert_called_with(['git', 'reset', '--soft', 'path'],
                               capture=False)

        git_client.reset(3, 'path')
        run.assert_called_with(['git', 'reset', '--hard', 'path'],
                               capture=False)

    @patch('rpkglib.GitClient.git_props', lambda: True)
    @patch('rpkglib.git_client.run')
    def test_status(self, run):
        git_client = GitClient(self.cwd)

        git_client.status()
        run.assert_called_with(['git', 'status'], capture=False)

        git_client.status(['pathspec1', 'pathspec2'])
        run.assert_called_with(['git', 'status', '--', 'pathspec1', 'pathspec2'],
                               capture=False)

    @patch('rpkglib.GitClient.git_props', lambda: True)
    @patch('rpkglib.GitClient.tag_message')
    @patch('rpkglib.git_client.run', silent_run)
    @patch('rpkglib.git_client.log', Mock())
    def test_add_tag(self, tag_message):
        self.init_repo()
        git_client = GitClient(self.cwd)
        tag_message.return_value = 'message'

        with pytest.raises(RpkgError):
            git_client.add_tag('foo')

        run(['git', 'tag', 'x-1-1'])
        with pytest.raises(RpkgError):
            git_client.add_tag('x-1-1')

        git_client.add_tag('y-1-1')
        result = run(['git', 'tag']).out.split()
        self.assertEqual(result, ['x-1-1', 'y-1-1'])

        with patch('rpkglib.git_client.run') as run_mock:
            git_client = GitClient(self.cwd, gpg_keyid='abc')
            git_client.add_tag('y-1-2')
            run_mock.assert_called_with(['git', 'tag', '-u', 'abc', '-m', 'message', 'y-1-2'])

    @patch('rpkglib.GitClient.git_props', lambda: True)
    @patch('rpkglib.git_client.run')
    def test_list_tags(self, run):
        git_client = GitClient(self.cwd)
        git_client.list_tags()
        run.assert_called_with(['git', 'tag', '-l'], capture=False)
        git_client.list_tags('tagname')
        run.assert_called_with(['git', 'tag', '-l', 'tagname'],
                               capture=False)

    @patch('rpkglib.GitClient.git_props', lambda: True)
    @patch('rpkglib.git_client.run')
    def test_delete_tag(self, run):
        git_client = GitClient(self.cwd)
        git_client.delete_tag('tagname')
        run.assert_called_with(['git', 'tag', '-d', 'tagname'],
                               capture=False)
